@@ -8,9 +8,14 @@ import struct
 from reversebox.common.logger import get_logger
 from reversebox.image.common import convert_bpp_to_bytes_per_pixel
 from reversebox.image.decoders.bumpmap_decoder import BumpmapDecoder
-from reversebox.image.decoders.compressed_decoder import CompressedImageDecoder
+from reversebox.image.decoders.compressed_decoder_encoder import (
+    CompressedImageDecoderEncoder,
+)
 from reversebox.image.decoders.gst_decoder import GSTImageDecoder
 from reversebox.image.decoders.n64_decoder import N64Decoder
+from reversebox.image.decoders.pvrtexlib_decoder_encoder import (
+    PvrTexlibImageDecoderEncoder,
+)
 from reversebox.image.decoders.yuv_decoder import YUVDecoder
 from reversebox.image.image_formats import ImageFormats
 from reversebox.image.pillow_wrapper import PillowWrapper
@@ -462,6 +467,15 @@ class ImageDecoder:
         p[3] = 0xFF
         return p
 
+    def _decode_rgba6666_pixel(self, pixel_int: int) -> bytes:
+        p = bytearray(4)
+        p[0] = (((pixel_int >> 0) & 63) * 255 + 32) // 63
+        p[1] = (((pixel_int >> 8) & 63) * 255 + 32) // 63
+        p[2] = (((pixel_int >> 16) & 63) * 255 + 32) // 63
+        a = (((pixel_int >> 24) & 63) * 255 + 32) // 63
+        p[3] = 0xFF if a else 0x00
+        return p
+
     def _decode_rgb48_pixel(self, pixel_int: int) -> bytes:
         p = bytearray(4)
         r = (pixel_int & 0xFFFF)
@@ -595,6 +609,8 @@ class ImageDecoder:
         ImageFormats.G16: (_decode_g_only_pixel, 16, get_uint16),
         ImageFormats.B16: (_decode_b_only_pixel, 16, get_uint16),
 
+        ImageFormats.RGBX6666: (_decode_rgbx6666_pixel, 24, get_uint24),
+        ImageFormats.RGBA6666: (_decode_rgba6666_pixel, 24, get_uint24),
         ImageFormats.RGB888: (_decode_rgb888_pixel, 24, get_uint24),
         ImageFormats.BGR888: (_decode_bgr888_pixel, 24, get_uint24),
 
@@ -792,8 +808,10 @@ class ImageDecoder:
         return self._decode_indexed(image_data, palette_data, img_width, img_height, image_format, palette_format, image_endianess, palette_endianess)
 
     def decode_compressed_image(self, image_data: bytes, img_width: int, img_height: int, image_format: ImageFormats) -> bytes:
-        compressed_decoder = CompressedImageDecoder()
-        return compressed_decoder.decode_compressed_image_main(image_data, img_width, img_height, image_format)
+        return CompressedImageDecoderEncoder().decode_compressed_image_main(image_data, img_width, img_height, image_format)
+
+    def decode_pvrtexlib_image(self, image_data: bytes, img_width: int, img_height: int, image_format: ImageFormats) -> bytes:
+        return PvrTexlibImageDecoderEncoder().decode_compressed_image_main(image_data, img_width, img_height, image_format)
 
     def decode_gst_image(self, image_data: bytes, palette_data: bytes, img_width: int, img_height: int, image_format: ImageFormats, convert_format: ImageFormats, convert_pal_format: ImageFormats, is_swizzled: bool = True) -> bytes:
         gst_decoder = GSTImageDecoder()
