@@ -5,11 +5,22 @@ License: GPL-3.0 License
 
 from typing import Dict, List, Tuple
 
+import PIL.Image
+from PIL import Image
+
 from reversebox.common.logger import get_logger
 from reversebox.image.common import (
     convert_bpp_to_bytes_per_pixel,
     get_bpp_for_image_format,
+    get_linear_image_data_size,
 )
+from reversebox.image.decoders.compressed_decoder_encoder import (
+    CompressedImageDecoderEncoder,
+)
+from reversebox.image.decoders.pvrtexlib_decoder_encoder import (
+    PvrTexlibImageDecoderEncoder,
+)
+from reversebox.image.image_decoder import ImageDecoder
 from reversebox.image.image_formats import ImageFormats
 from reversebox.image.pillow_wrapper import PillowWrapper
 from reversebox.io_files.bytes_handler import BytesHandler
@@ -144,6 +155,25 @@ class ImageEncoder:
         p[1] = (bgra4444 >> 8) & 0xFF
         return p
 
+    def _encode_rgbx4444_pixel(self, pixel_int: int) -> bytearray:
+        p = bytearray(2)
+
+        r = pixel_int & 0xFF
+        g = (pixel_int >> 8) & 0xFF
+        b = (pixel_int >> 16) & 0xFF
+        x = 0xFF
+
+        r4 = (r >> 4) & 0x0F
+        g4 = (g >> 4) & 0x0F
+        b4 = (b >> 4) & 0x0F
+        x4 = (x >> 4) & 0x0F
+
+        rgbx4444 = (x4 << 12) | (b4 << 8) | (g4 << 4) | r4
+
+        p[0] = rgbx4444 & 0xFF
+        p[1] = (rgbx4444 >> 8) & 0xFF
+        return p
+
     def _encode_rgba5551_pixel(self, pixel_int: int) -> bytearray:
         p = bytearray(2)
 
@@ -234,6 +264,7 @@ class ImageEncoder:
         ImageFormats.BGR565: (_encode_bgr565_pixel, 16, get_uint16, set_uint16),
         ImageFormats.ABGR4444: (_encode_abgr4444_pixel, 16, get_uint16, set_uint16),
         ImageFormats.BGRA4444: (_encode_bgra4444_pixel, 16, get_uint16, set_uint16),
+        ImageFormats.RGBX4444: (_encode_rgbx4444_pixel, 16, get_uint16, set_uint16),
         ImageFormats.RGBA5551: (_encode_rgba5551_pixel, 16, get_uint16, set_uint16),
         ImageFormats.BGRA5551: (_encode_bgra5551_pixel, 16, get_uint16, set_uint16),
         ImageFormats.RGBX5551: (_encode_rgbx5551_pixel, 16, get_uint16, set_uint16),
@@ -254,7 +285,7 @@ class ImageEncoder:
             raise Exception("Wrong endianess!")
         return endianess_format
 
-    def _encode_generic(self, image_data: bytes, img_width: int, img_height: int, image_format: ImageFormats, image_endianess: str) -> bytes:
+    def _encode_generic_image(self, image_data: bytes, img_width: int, img_height: int, image_format: ImageFormats, image_endianess: str) -> bytes:
         encode_function, bits_per_pixel, _, _ = self.generic_data_formats[image_format]
         image_handler = BytesHandler(image_data)
         source_image_bytes_per_pixel = 4  # always 4 for RGBA8888
@@ -295,57 +326,141 @@ class ImageEncoder:
 
         return texture_data
 
+    def _encode_generic(self, image_data: bytes, img_width: int, img_height: int, image_format: ImageFormats,
+                        image_endianess: str, number_of_mipmaps: int, mipmaps_resampling_type: PIL.Image.Resampling) -> bytes:
+        # main image logic
+        encoded_texture_data: bytes = self._encode_generic_image(image_data, img_width, img_height, image_format, image_endianess)
+
+        # mipmaps logic
+        if number_of_mipmaps > 0:
+            base_img: Image = PillowWrapper().get_pillow_image_from_rgba8888_data(
+                image_data, img_width, img_height
+            )
+            mip_width: int = img_width
+            mip_height: int = img_height
+            for i in range(number_of_mipmaps):
+                mip_width //= 2
+                mip_height //= 2
+                mip_pillow_img: Image = base_img.resize((mip_width, mip_height), resample=mipmaps_resampling_type)
+                mip_rgba_data: bytes = PillowWrapper().get_image_data_from_pillow_image(mip_pillow_img)
+                encoded_mipmap_data: bytes = self._encode_generic_image(mip_rgba_data, mip_width, mip_height, image_format, image_endianess)
+                encoded_texture_data += encoded_mipmap_data
+
+        return encoded_texture_data
+
+    def _encode_indexed_get_intermediate_image(self, image_data: bytes, img_width: int, img_height: int,
+                                               palette_format: ImageFormats, image_endianess: str, max_colors_count: int) -> bytes:
+
+        # colour quantization and dithering
+        pillow_img: Image = PillowWrapper().get_pillow_image_from_rgba8888_data(image_data, img_width, img_height)
+        unique_colors_count: int = len(set(pillow_img.getdata()))
+        if unique_colors_count > max_colors_count:
+            pillow_img = pillow_img.quantize(colors=max_colors_count, method=Image.Quantize.FASTOCTREE).convert("RGBA", dither=Image.Dither.FLOYDSTEINBERG)
+        image_data = pillow_img.tobytes()
+
+        # encode to intermediate image (e.g. RGBA8888 -> RGB565)
+        return self._encode_generic(image_data, img_width, img_height, palette_format, image_endianess, 0, PIL.Image.Resampling.NEAREST)
+
+    def _encode_indexed_revert_intermediate_image(self, image_data: bytes, img_width: int, img_height: int,
+                                                  palette_format: ImageFormats, image_endianess: str) -> bytes:
+
+        # revert intermediate image (e.g. RGB565 -> RGBA8888)
+        return ImageDecoder().decode_image(image_data, img_width, img_height, palette_format, image_endianess)
+
+    def _encode_indexed_calculate_output_size(self, image_bpp: int, img_width: int, img_height: int) -> int:
+        if image_bpp == 4:
+            return img_width * img_height // 2
+        elif image_bpp == 8:
+            return img_width * img_height
+        elif image_bpp == 16:
+            return img_width * img_height * 2
+        elif image_bpp == 24:
+            return img_width * img_height * 3
+        elif image_bpp == 32:
+            return img_width * img_height * 4
+        else:
+            raise Exception(f"[1] Image_bits_per_pixel={image_bpp} not supported!")
+
+    def _encode_indexed_encode_indices(self, image_bpp: int, pixel_int_values: list[int], pixel_map: Dict[int, int], palette_endianess: str,
+                                       image_endianess_format: str, image_bytes_per_pixel: int, texture_data: bytes):
+        img_entry_number: int = 0
+        if image_bpp == 4:  # PAL4
+            for i in range(0, len(pixel_int_values), 2):
+                pal_entry_number_1: int = pixel_map[pixel_int_values[i]]
+                pal_entry_number_2: int = pixel_map[pixel_int_values[i+1]]
+                if palette_endianess == "little":
+                    pal_entry_combined: int = ((pal_entry_number_2 << 4) | pal_entry_number_1) & 0xFF
+                else:
+                    pal_entry_combined: int = ((pal_entry_number_1 << 4) | pal_entry_number_2) & 0xFF
+                pixel_bytes: bytes = set_uint8(pal_entry_combined, image_endianess_format)
+                texture_data[img_entry_number * image_bytes_per_pixel: (img_entry_number + 1) * image_bytes_per_pixel] = pixel_bytes
+                img_entry_number += 1
+
+        elif image_bpp == 8:  # PAL8
+            for pixel_int_value in pixel_int_values:
+                pal_entry_number: int = pixel_map[pixel_int_value]
+                pixel_bytes: bytes = set_uint8(pal_entry_number, image_endianess_format)
+                texture_data[img_entry_number * image_bytes_per_pixel: (img_entry_number + 1) * image_bytes_per_pixel] = pixel_bytes
+                img_entry_number += 1
+
+        else:
+            raise Exception(f"Not supported img_bpp={image_bpp}!")  # TODO - support other formats like PAL16
+
+        return texture_data
+
+    def _encode_indexed_get_pixel_int_values(self, image_bpp: int, encoded_intermediate_image: bytes, palette_bytes_per_pixel: int,
+                                             read_function, image_endianess_format: str) -> list[int]:
+        pixel_int_values: list[int] = []
+        if image_bpp in (4, 8):
+            number_of_pixels: int = len(encoded_intermediate_image) // palette_bytes_per_pixel
+            for i in range(number_of_pixels):
+                pixel_bytes: bytes = encoded_intermediate_image[
+                                     i * palette_bytes_per_pixel: (i + 1) * palette_bytes_per_pixel]
+                pixel_int: int = read_function(pixel_bytes, image_endianess_format)
+                pixel_int_values.append(pixel_int)
+        else:
+            raise Exception(f"Not supported img_bpp={image_bpp}!")  # TODO - support other formats like PAL16
+
+        return pixel_int_values
+
     def _encode_indexed(self, image_data: bytes, img_width: int,
-                        img_height: int, image_format: ImageFormats, palette_format: ImageFormats, image_endianess: str, palette_endianess: str) -> Tuple[bytes, bytes]:
+                        img_height: int, image_format: ImageFormats, palette_format: ImageFormats,
+                        image_endianess: str, palette_endianess: str, max_colors_count: int, number_of_mipmaps: int) -> Tuple[bytes, bytes]:
         encode_function, palette_bpp, read_function, write_function = self.generic_data_formats[palette_format]
         image_bpp: int = get_bpp_for_image_format(image_format)
         palette_bpp: int = get_bpp_for_image_format(palette_format)
         palette_bytes_per_pixel: int = convert_bpp_to_bytes_per_pixel(palette_bpp)
         image_bytes_per_pixel: int = convert_bpp_to_bytes_per_pixel(image_bpp)
         image_endianess_format: str = self._get_endianess_format(image_endianess)
+        texture_data: bytes = b''
 
         # prepare empty bytearray for output data
-        if image_bpp == 4:
-            texture_data = bytearray(img_width * img_height // 2)
-        elif image_bpp == 8:
-            texture_data = bytearray(img_width * img_height)
-        elif image_bpp == 16:
-            texture_data = bytearray(img_width * img_height * 2)
-        elif image_bpp == 24:
-            texture_data = bytearray(img_width * img_height * 3)
-        elif image_bpp == 32:
-            texture_data = bytearray(img_width * img_height * 4)
-        else:
-            raise Exception(f"[1] Image_bits_per_pixel={image_bpp} not supported!")
+        main_texture_data = bytearray(self._encode_indexed_calculate_output_size(image_bpp, img_width, img_height))
+        main_texture_data_expected_size: int = len(main_texture_data)
 
-        # encode to intermediate image
-        encoded_intermediate_image: bytes = self._encode_generic(image_data, img_width, img_height, palette_format, image_endianess)  # RGBA8888
+        # get intermediate image
+        encoded_intermediate_image: bytes = self._encode_indexed_get_intermediate_image(image_data, img_width, img_height,
+                                                                                        palette_format, image_endianess, max_colors_count)
+        expected_encoded_intermediate_image_size: int = get_linear_image_data_size(get_bpp_for_image_format(palette_format), img_width, img_height)
+        if len(encoded_intermediate_image) != expected_encoded_intermediate_image_size:
+            raise Exception("Error! Wrong intermediate image data size!")
 
         # get pixels from intermediate image
-        pixel_int_values: list[int] = []
-        if image_bpp in (4, 8):
-            number_of_pixels: int = len(encoded_intermediate_image) // palette_bytes_per_pixel
-            for i in range(number_of_pixels):
-                pixel_bytes: bytes = encoded_intermediate_image[i * palette_bytes_per_pixel: (i + 1) * palette_bytes_per_pixel]
-                pixel_int: int = read_function(pixel_bytes, image_endianess_format)
-                pixel_int_values.append(pixel_int)
+        pixel_int_values: list[int] = self._encode_indexed_get_pixel_int_values(image_bpp, encoded_intermediate_image, palette_bytes_per_pixel,
+                                                                                read_function, image_endianess_format)
+
+        # get aligned number of colors
+        if max_colors_count <= 4:  # PAL2 (4-color)
+            aligned_colors_count = 4
+        elif max_colors_count <= 16:  # PAL4 (16-color)
+            aligned_colors_count = 16
+        elif max_colors_count <= 256:  # PAL8 (256-color)
+            aligned_colors_count = 256
         else:
-            raise Exception(f"Not supported img_bpp={image_bpp}!")  # TODO - support other formats like PAL16
+            raise Exception(f"Not supported aligned colors count for max_colors_count={max_colors_count}")  # TODO - support other formats like PAL16
 
-        # get number of unique colors
-        unique_colors_count: int = len(set(pixel_int_values))
-
-        # get max number of colors
-        if unique_colors_count <= 4:  # PAL2 (4-color)
-            max_colors_count = 4
-        elif unique_colors_count <= 16:  # PAL4 (16-color)
-            max_colors_count = 16
-        elif unique_colors_count <= 256:  # PAL8 (256-color)
-            max_colors_count = 256
-        else:
-            raise Exception(f"Not supported max colors count for unique_colors={unique_colors_count}")  # TODO - support other formats like PAL16
-
-        palette_data: bytearray = bytearray(max_colors_count * palette_bytes_per_pixel)
+        palette_data: bytearray = bytearray(aligned_colors_count * palette_bytes_per_pixel)
+        palette_data_expected_size: int = len(palette_data)
 
         # palette preparation (get unique values for palette)
         pixel_check_list: List[int] = []
@@ -364,32 +479,64 @@ class ImageEncoder:
             palette_data[pal_entry_number * palette_bytes_per_pixel: (pal_entry_number + 1) * palette_bytes_per_pixel] = pal_entry_bytes
             pal_entry_number += 1
 
-        # encode indices
-        img_entry_number: int = 0
+        # encode indices for main texture
+        main_texture_data: bytes = self._encode_indexed_encode_indices(image_bpp, pixel_int_values, pixel_map, palette_endianess,
+                                                                       image_endianess_format, image_bytes_per_pixel, main_texture_data)
+        texture_data += main_texture_data
 
-        if image_bpp == 4:  # PAL4
-            for i in range(0, len(pixel_int_values), 2):
-                pal_entry_number_1: int = pixel_map[pixel_int_values[i]]
-                pal_entry_number_2: int = pixel_map[pixel_int_values[i+1]]
-                pal_entry_combined: int = (pal_entry_number_1 << 4) | pal_entry_number_2
-                pixel_bytes: bytes = set_uint8(pal_entry_combined, image_endianess_format)
-                texture_data[img_entry_number * image_bytes_per_pixel: (img_entry_number + 1) * image_bytes_per_pixel] = pixel_bytes
-                img_entry_number += 1
+        # mipmaps logic
+        if number_of_mipmaps > 0:
+            base_rgba_data: bytes = self._encode_indexed_revert_intermediate_image(encoded_intermediate_image, img_width, img_height,
+                                                                                   palette_format, image_endianess)
+            base_img: Image = PillowWrapper().get_pillow_image_from_rgba8888_data(base_rgba_data, img_width, img_height)
+            mip_width: int = img_width
+            mip_height: int = img_height
+            for i in range(number_of_mipmaps):
+                mip_width //= 2
+                mip_height //= 2
+                mip_pillow_img: Image = base_img.resize((mip_width, mip_height), resample=PIL.Image.Resampling.NEAREST)
+                mip_rgba_data: bytes = PillowWrapper().get_image_data_from_pillow_image(mip_pillow_img)
+                mipmap_intermediate_image: bytes = self._encode_indexed_get_intermediate_image(mip_rgba_data, mip_width,
+                                                                                               mip_height, palette_format,
+                                                                                               image_endianess, max_colors_count)
+                mipmap_pixel_int_values: list[int] = self._encode_indexed_get_pixel_int_values(image_bpp,
+                                                                                               mipmap_intermediate_image,
+                                                                                               palette_bytes_per_pixel,
+                                                                                               read_function,
+                                                                                               image_endianess_format)
 
-        elif image_bpp == 8:  # PAL8
-            for pixel_int_value in pixel_int_values:
-                pal_entry_number: int = pixel_map[pixel_int_value]
-                pixel_bytes: bytes = set_uint8(pal_entry_number, image_endianess_format)
-                texture_data[img_entry_number * image_bytes_per_pixel: (img_entry_number + 1) * image_bytes_per_pixel] = pixel_bytes
-                img_entry_number += 1
+                mipmap_texture_data = bytearray(self._encode_indexed_calculate_output_size(image_bpp, mip_width, mip_height))
+                mipmap_texture_data_expected_size: int = len(mipmap_texture_data)
+                mipmap_texture_data: bytes = self._encode_indexed_encode_indices(image_bpp, mipmap_pixel_int_values, pixel_map, palette_endianess,
+                                                                                 image_endianess_format, image_bytes_per_pixel, mipmap_texture_data)
 
-        else:
-            raise Exception(f"Not supported img_bpp={image_bpp}!")  # TODO - support other formats like PAL16
+                if len(mipmap_texture_data) != mipmap_texture_data_expected_size:
+                    raise Exception("Error! Wrong size of mipmap data!")
+
+                texture_data += mipmap_texture_data
+
+        # final checks
+        if len(palette_data) != palette_data_expected_size or len(palette_data) > 1024:
+            raise Exception("Error! Wrong size of palette data!")
+
+        if len(main_texture_data) != main_texture_data_expected_size:
+            raise Exception("Error! Wrong size of main texture data!")
 
         return texture_data, palette_data
 
-    def encode_image(self, image_data: bytes, img_width: int, img_height: int, image_format: ImageFormats, image_endianess: str = "little") -> bytes:
-        return self._encode_generic(image_data, img_width, img_height, image_format, image_endianess)
+    def encode_image(self, image_data: bytes, img_width: int, img_height: int, image_format: ImageFormats,
+                     image_endianess: str = "little", number_of_mipmaps: int = 0,
+                     mipmaps_resampling_type: PIL.Image.Resampling = Image.Resampling.NEAREST) -> bytes:
+        return self._encode_generic(image_data, img_width, img_height, image_format, image_endianess, number_of_mipmaps, mipmaps_resampling_type)
 
-    def encode_indexed_image(self, image_data: bytes, img_width: int, img_height: int, image_format: ImageFormats, palette_format: ImageFormats, image_endianess: str = "little", palette_endianess: str = "little") -> Tuple[bytes, bytes]:
-        return self._encode_indexed(image_data, img_width, img_height, image_format, palette_format, image_endianess, palette_endianess)
+    def encode_indexed_image(self, image_data: bytes, img_width: int, img_height: int, image_format: ImageFormats,
+                             palette_format: ImageFormats, max_color_count: int, image_endianess: str = "little",
+                             palette_endianess: str = "little", number_of_mipmaps: int = 0) -> Tuple[bytes, bytes]:
+        return self._encode_indexed(image_data, img_width, img_height, image_format, palette_format,
+                                    image_endianess, palette_endianess, max_color_count, number_of_mipmaps)
+
+    def encode_compressed_image(self, image_data: bytes, img_width: int, img_height: int, image_format: ImageFormats) -> bytes:
+        return CompressedImageDecoderEncoder().encode_compressed_image_main(image_data, img_width, img_height, image_format)
+
+    def encode_pvrtexlib_image(self, image_data: bytes, img_width: int, img_height: int, image_format: ImageFormats) -> bytes:
+        return PvrTexlibImageDecoderEncoder().encode_compressed_image_main(image_data, img_width, img_height, image_format)
